from .pyplot import interactive_axhline as axhline
from .pyplot import interactive_axvline as axvline
from .pyplot import interactive_hist as hist
from .pyplot import interactive_imshow as imshow
from .pyplot import interactive_plot as plot
from .pyplot import interactive_scatter as scatter
from .pyplot import interactive_text as text
from .pyplot import interactive_title as title
from .pyplot import interactive_xlabel as xlabel
from .pyplot import interactive_ylabel as ylabel
